@@ -5,11 +5,8 @@ from enum import IntEnum
 
 import keras
 import numpy as np
-from keras.layers import *
-from keras.losses import mean_squared_error
-from keras import backend as K
 
-from landlordai.game.card import Card
+from landlordai.game.card import Card, string_to_card
 from landlordai.game.deck import CardSet
 from landlordai.game.move import KittyReveal, SpecificMove, BetMove
 
@@ -44,13 +41,16 @@ class Player:
     def make_move(self, game):
         pass
 
+    def get_name(self):
+        return self.name
+
+
 class LearningPlayer(Player):
     TIMESTEPS = 100
 
     #estimation methods for q(s', a')
-    MONTECARLO_RANDOM = 'mcrandom'
-    BEST_SIMULATION = 'bestsim'
-    CONSENSUS_Q = 'consensusq'
+    NO_ESTIMATION = 'no_estimation'
+    ACTUAL_Q = 'actualq'
 
     # 12: number of distinct cards, each feature is the number played
     # 3: one-hot encoding for landlordai player
@@ -62,16 +62,16 @@ class LearningPlayer(Player):
     HAND_FEATURES = len(Card) + 3
 
     def __init__(self, name, net_dir=None, epsilon=0.1, learning_rate=0.2, discount_factor=0.8,
-                  estimation_mode=MONTECARLO_RANDOM,
+                  estimation_mode=ACTUAL_Q,
                  random_mc_num_explorations=30,
-                 mc_best_move_depth=1):
+                 estimation_depth=1):
         super().__init__(name)
 
         self.epsilon = epsilon
         self.empty_nets = False
         self.estimation_mode = estimation_mode
         self.random_mc_num_explorations = random_mc_num_explorations
-        self.mc_best_move_depth = mc_best_move_depth
+        self.estimation_depth = estimation_depth
         if net_dir is None:
             self.empty_nets = True
         else:
@@ -91,6 +91,7 @@ class LearningPlayer(Player):
         self.record_hand_vectors = []
         # for debugging
         self._record_state_q = []
+        self._recording_finalized = False
 
     def load_nnets(self, net_dir):
         self.history_net = keras.models.load_model(net_dir + "/history.h5")
@@ -134,7 +135,21 @@ class LearningPlayer(Player):
         return self.feature_index[i]
 
     def _derive_move_stack(self, game):
-        return np.vstack([self.compute_move_vector(player, game.get_landlord_position(), move) for player, move in game.get_game_logs()])
+        # pad to achieve timestep length
+        feature_stack = []
+        move_stack = [self.compute_move_vector(player, game.get_landlord_position(), move) for player, move in game.get_game_logs()]
+        feature_stack.extend(move_stack)
+        if len(feature_stack) == 0:
+            return []
+        return np.vstack(feature_stack)
+
+    def _derive_move_stack_bridge(self, game, player):
+        # pad to achieve timestep length
+        feature_stack = [np.concatenate([self.get_hand_vector(game, player),
+                                         np.zeros(LearningPlayer.TIMESTEP_FEATURES - LearningPlayer.HAND_FEATURES)])]
+        move_stack = [self.compute_move_vector(player, game.get_landlord_position(), move) for player, move in game.get_game_logs()]
+        feature_stack.extend(move_stack)
+        return np.vstack(feature_stack)
 
     def _append_padding(self, move_stack):
         fluff_volume = LearningPlayer.TIMESTEPS - len(move_stack)
@@ -148,24 +163,10 @@ class LearningPlayer(Player):
         return np.vstack([move_stack, fluff_stack])
 
     def derive_features(self, game):
-        if len(game.get_game_logs()) <= 0:
-            move_stack = []
-        else:
-            move_stack = self._derive_move_stack(game)
-        return self._append_padding(move_stack)
+        return self._append_padding(self._derive_move_stack(game))
 
-    # separately returns current game state - 1 = history_matrix, and separate move vector
-    def derive_record_features(self, game):
-        move_vector = np.zeros(LearningPlayer.TIMESTEP_FEATURES)
-        if len(game.get_game_logs()) == 1:
-            last_player, last_move = game.get_game_logs()[0]
-            move_vector = self.compute_move_vector(last_player, game.get_landlord_position(), last_move)
-        if len(game.get_game_logs()) <= 1:
-            history_matrix = self._append_padding([])
-            return history_matrix, move_vector
-
-        move_stack = self._derive_move_stack(game)
-        return self._append_padding(move_stack[:-1]), move_stack[-1]
+    def derive_features_bridge(self, game, player: TurnPosition):
+        return self._append_padding(self._derive_move_stack_bridge(game, player))
 
     def compute_move_vector(self, player: TurnPosition, landlord_position: TurnPosition, move):
         move_vector = np.zeros(LearningPlayer.TIMESTEP_FEATURES)
@@ -190,9 +191,6 @@ class LearningPlayer(Player):
         if type(move) == SpecificMove:
             for card, count in move.cards.items():
                 move_vector[self.get_feature_index(card.name)] = count
-            # print(1 if game.current_player_is_landlord() else 0)
-            # print(1 if game.get_current_position().next() == game.get_landlord_position() else 0)
-            # print(1 if game.get_current_position().previous() == game.get_landlord_position() else 0)
             other_features = {'I_AM_LANDLORD': 1 if player == landlord_position else 0,
                               'I_AM_BEFORE_LANDLORD': 1 if player.previous() == landlord_position else 0,
                               'I_AM_AFTER_LANDLORD': 1 if player.next() == landlord_position else 0}
@@ -221,13 +219,14 @@ class LearningPlayer(Player):
         if self.empty_nets:
             return np.random.random(LearningPlayer.TIMESTEP_FEATURES) * 0.01
 
-        return self.history_net.predict(np.array([features]), batch_size=1024)[0]
+        return self.history_net.predict(np.array([features]), batch_size=1)[0]
 
     def get_position_predictions(self, history_matrix, move_options_matrix, hand_matrix):
         if self.empty_nets:
             return np.random.random((move_options_matrix.shape[0])) * 0.01
 
-        return self.position_net.predict([history_matrix, move_options_matrix, hand_matrix], batch_size=1024).reshape(move_options_matrix.shape[0])
+        num_rows = move_options_matrix.shape[0]
+        return self.position_net.predict([history_matrix, move_options_matrix, hand_matrix], batch_size=num_rows).reshape(num_rows)
 
     def make_bet_decision(self, game, legal_moves, predictions):
         bet_indices = []
@@ -240,10 +239,7 @@ class LearningPlayer(Player):
 
         return np.argmax(flip_predictions)
 
-    def decide_best_move(self, game, debug=False):
-        assert len(game.get_hand(game.get_current_position())) > 0
-        legal_moves = game.get_legal_moves()
-
+    def full_move_evaluation(self, game, legal_moves):
         history_features = self.derive_features(game)
 
         # all the moves we make from here will not affect the history, so assess it and copy
@@ -260,6 +256,14 @@ class LearningPlayer(Player):
 
         predictions = self.get_position_predictions(history_matrix, move_options_matrix, hand_matrix)
 
+        return predictions
+
+    def decide_best_move(self, game, debug=False):
+        assert len(game.get_hand(game.get_current_position())) > 0
+        legal_moves = game.get_legal_moves()
+
+        predictions = self.full_move_evaluation(game, legal_moves)
+
         # for debugging
         raw_predictions = np.copy(predictions)
 
@@ -305,80 +309,34 @@ class LearningPlayer(Player):
     def make_move(self, game, debug=False):
         best_move, best_move_q = self.decide_best_move(game, debug=debug)
 
-        self.record_move(game, best_move, best_move_q, game.get_current_position())
+        if self.estimation_mode == LearningPlayer.ACTUAL_Q:
+            self.record_move(game, best_move, best_move_q, game.get_current_position())
 
         return best_move
 
-
-
-    # returns a future q based on random search
-    def monte_carlo_best_search(self, game, depth=6):
-        copy_game = copy(game)
-        best_next_move_q = 0
-        for i in range(depth):
-            best_next_move, best_next_move_q = self.decide_best_move(copy_game)
-            if copy_game.move_ends_game(best_next_move):
-                return best_next_move_q
-            else:
-                copy_game.play_move(best_next_move)
-
-        return best_next_move_q
-
-    def monte_carlo_random_search(self, game, num_explorations):
-        reward_values = []
-        for i in range(num_explorations):
-            copy_game = copy(game)
-            while not copy_game.is_round_over():
-                move = random.choice(copy_game.get_legal_moves())
-                copy_game.play_move(move)
-            reward_values.append(copy_game.get_r())
-        return np.mean(reward_values)
-
-    def consensus_q(self, game):
-        copy_game = copy(game)
-        best_next_move_qs = []
-        # one for each player
-        for i in range(3):
-            best_next_move, best_next_move_q = self.decide_best_move(copy_game)
-            best_next_move_qs.append(best_next_move_q)
-            if copy_game.move_ends_game(best_next_move):
-                return best_next_move_q
-            else:
-                copy_game.play_move(best_next_move)
-
-        return np.mean(best_next_move_qs)
-
     def record_move(self, game, best_move, best_move_q, player: TurnPosition):
-        #previous_history_matrix, move_vector = self.derive_record_features(game)
-        history_matrix = self.derive_features(game)
+        history_matrix = self.derive_features_bridge(game, player)
         move_vector = self.compute_move_vector(player, game.get_landlord_position(), best_move)
         hand_vector = self.get_hand_vector(game, player)
 
-        copy_game = copy(game)
-        copy_game.play_move(best_move)
-        immediate_reward = 0
-        future_reward = 0
-        if game.move_ends_game(best_move):
-            immediate_reward = copy_game.get_r()
-        else:
-            if self.estimation_mode == LearningPlayer.MONTECARLO_RANDOM:
-                future_reward = self.monte_carlo_random_search(copy_game, num_explorations=self.random_mc_num_explorations)
-            elif self.estimation_mode == LearningPlayer.BEST_SIMULATION:
-                future_reward = self.monte_carlo_best_search(copy_game, depth=self.mc_best_move_depth)
-            else:
-                future_reward = self.consensus_q(copy_game)
-
         self.record_history_matrices.append(history_matrix)
         self.record_move_vectors.append(move_vector)
         self.record_hand_vectors.append(hand_vector)
 
-        # take old value and move by the newly learned value
-        q_update = best_move_q + self.learning_rate * (immediate_reward + self.discount_factor * future_reward - best_move_q)
-        #if abs(int(q_update) - q_update) < 1E-5:
-        #    print('Diff')
-
         self._record_state_q.append(best_move_q)
-        self._record_future_q.append(q_update)
+
+    def compute_future_q(self, game):
+        assert self._recording_finalized is False
+
+        self._record_state_q.append(game.get_r())
+        # shift forward
+        self._record_future_q = []
+        for i, experienced_q in enumerate(self._record_state_q[1:]):
+            old_q = self._record_state_q[i]
+            update_q = old_q + self.learning_rate * (self.discount_factor * experienced_q - old_q)
+            self._record_future_q.append(update_q)
+
+        self._recording_finalized = True
 
     def get_record_history_matrices(self):
         return self.record_history_matrices
@@ -389,18 +347,42 @@ class LearningPlayer(Player):
     def get_record_hand_vectors(self):
         return self.record_hand_vectors
 
-    def get_future_q(self):
+    def get_estimated_qs(self):
+        assert self._recording_finalized is True
         return self._record_future_q
 
-    def get_name(self):
-        return self.name
-
     def __str__(self):
         return self.get_name()
 
     def __repr__(self):
         return self.__str__()
 
+
+class LearningPlayer_v2(LearningPlayer):
+    def _derive_feature_stack(self, game, player):
+        # pad to achieve timestep length
+        feature_stack = [np.concatenate([self.get_hand_vector(game, player),
+                                         np.zeros(LearningPlayer.TIMESTEP_FEATURES - LearningPlayer.HAND_FEATURES)])]
+        move_stack = [self.compute_move_vector(player, game.get_landlord_position(), move) for player, move in game.get_game_logs()]
+        feature_stack.extend(move_stack)
+        return np.vstack(feature_stack)
+
+    def derive_features(self, game):
+        return self._append_padding(self._derive_feature_stack(game, game.get_current_position()))
+
+    def record_move(self, game, best_move, best_move_q, player: TurnPosition):
+        history_matrix = self.derive_features(game)
+
+        move_vector = self.compute_move_vector(player, game.get_landlord_position(), best_move)
+        hand_vector = self.get_hand_vector(game, player)
+
+        self.record_history_matrices.append(history_matrix)
+        self.record_move_vectors.append(move_vector)
+        self.record_hand_vectors.append(hand_vector)
+
+        self._record_state_q.append(best_move_q)
+
+
 class RandomPlayer(Player):
     def make_move(self, game, debug=False):
         legal_moves = game.get_legal_moves()
@@ -408,4 +390,111 @@ class RandomPlayer(Player):
 
 class NoBetPlayer(Player):
     def make_move(self, game, debug=False):
-        return None
\ No newline at end of file
+        return None
+
+
+class InvalidMoveError(Exception):
+    pass
+
+
+class TypoError(Exception):
+    pass
+
+
+class HumanPlayer(Player):
+    def __init__(self, name, reference_player=None):
+        super().__init__(name)
+        self.reference_player = reference_player
+        self.first_turn = True
+
+    def ask_ai(self, game, my_move, top_n=5):
+        legal_moves = game.get_legal_moves()
+        is_landlord = game.get_landlord_position() == game.get_current_position()
+        predictions = self.reference_player.full_move_evaluation(game, legal_moves)
+
+        sorted_moves = sorted(list(zip(legal_moves, predictions)), key=lambda x : -x[1] if is_landlord else x[1])
+
+        print("\tEvaluation By", self.reference_player.get_name())
+        if my_move == sorted_moves[0][0]:
+            print("\tBest Move!")
+
+        for i, move_tup in enumerate(sorted_moves):
+            move, pred = move_tup
+            move_marker = '<-----' if move == my_move else ''
+            if i < top_n or move == my_move:
+                print('\t', i, ": ", pred, move, move_marker)
+
+        print('\n')
+
+    @classmethod
+    def cardstrings_to_cards(cls, card_strings):
+        cards = []
+        try:
+            for card_string in card_strings:
+                card_string = card_string.strip()
+                try:
+                    card_string = card_string.upper()
+                except Exception:
+                    pass
+                cards.append(string_to_card(card_string))
+        except ValueError:
+            raise TypoError
+        return cards
+
+    @classmethod
+    def parse_input(cls, input_string, bet_phase=False):
+        if len(input_string) == 0:
+            return None
+
+        if bet_phase:
+            try:
+                bet_amount = int(input_string)
+                return BetMove(bet_amount)
+            except Exception:
+                pass
+
+        try:
+            card_strings = input_string.split(' ')
+        except Exception:
+            raise TypoError
+
+        cards = HumanPlayer.cardstrings_to_cards(card_strings)
+
+        all_possible_moves = CardSet(Counter(cards)).get_all_moves()
+        for move in all_possible_moves:
+            if move.get_cards() == Counter(cards):
+                return move
+        raise InvalidMoveError
+
+    def print_instructions(self):
+        if self.first_turn:
+            print("\nYour turn!")
+            print("Enter your cards separated by spaces. Example \"EIGHT EIGHT\".")
+            print("Enter a number to make a bet (between 0 and 3) or Return to pass.")
+            print("\n")
+            self.first_turn = False
+            self.print_instructions()
+        else:
+            print('\n')
+
+
+    def decide_best_move(self, game):
+        self.print_instructions()
+        while True:
+
+            print(game.get_hand(game.get_current_position()))
+            inp = input(">").strip()
+            try:
+                human_move = HumanPlayer.parse_input(inp, bet_phase=not game.is_betting_complete())
+                if human_move in game.get_legal_moves():
+                    self.ask_ai(game, human_move)
+                    return human_move, 0
+                else:
+                    print('Illegal Move, please try again!')
+            except TypoError:
+                print("Invalid Input, check your spelling!")
+            except InvalidMoveError:
+                print("Not a valid Landlord Move!")
+
+
+
